"""Various indexers for common formats"""
import logging
import datetime
import io
import zipfile
import re

import multidict

try:
    import pyexiv2
except ImportError:
    pyexiv2 = None

try:
    import PIL
    import PIL.Image
except ImportError:
    PIL = None

try:
    import mutagen
except ImportError:
    mutagen = None

try:
    import pdfminer
    from pdfminer.pdfparser import PDFParser
    from pdfminer.pdfdocument import PDFDocument
except ImportError:
    pdfminer = None

from metaindex.indexer import registered_indexer, to_utf8, Indexer, Order
from metaindex.opf import parse_opf
from metaindex.ruleindexer import RuleIndexer


if pyexiv2 is not None:
    @registered_indexer
    class Exiv2Indexer(Indexer):
        NAME = 'exiv2'
        ACCEPT = ['image/', 'video/']

        def run(self, path, info):
            logging.debug(f"[image, exiv2] processing {path.name}")

            try:
                meta = pyexiv2.core.Image(str(path))
            except:
                return False, {}

            result = multidict.MultiDict()
            try:
                result.extend(meta.read_exif())
            except:
                pass

            try:
                result.extend(meta.read_iptc())
            except:
                pass

            try:
                result.extend(meta.read_xmp())
            except:
                pass

            meta.close()
            return True, result


if mutagen is not None:
    @registered_indexer
    class MutagenIndexer(Indexer):
        NAME = 'mutagen'
        ACCEPT = ['audio/', 'video/']

        def run(self, path, info):
            logging.debug(f"[mutagen] processing {path.name}")

            try:
                meta = mutagen.File(path, easy=True)
            except:
                return False, {}

            result = multidict.MultiDict()

            if meta is not None:
                for key in meta.keys():
                    result.extend([('id3.' + key, value) for value in meta[key]])
                if hasattr(meta, 'info') and hasattr(meta.info, 'length'):
                    result.add('length', meta.info.length)

            return True, result


if PIL is not None:
    @registered_indexer
    class PillowIndexer(Indexer):
        NAME = 'pillow'
        ACCEPT = ['image/']

        def run(self, path, info):
            logging.debug(f"[image, pillow] processing {path.name}")

            try:
                meta = PIL.Image.open(path)
            except:
                return False, {}

            result = {}
            if meta is not None:
                result = {'resolution': "{}x{}".format(*meta.size)}
                if self.should_ocr(path):
                    ocr = self.ocr.run(meta)
                    if ocr.success:
                        result['ocr.language'] = ocr.language
                        if self.should_fulltext(path):
                            result['ocr.fulltext'] = ocr.fulltext
                meta.close()

            return True, result


if pdfminer is not None:
    @registered_indexer
    class PdfMinerIndexer(Indexer):
        NAME = 'pdfminer'
        ACCEPT = ['application/pdf']

        PDF_METADATA = ('title', 'author', 'creator', 'producer', 'keywords',
                        'manager', 'status', 'category', 'moddate', 'creationdate',
                        'subject')

        def run(self, path, info):
            logging.debug(f"[pdf] processing {path.name}")

            try:
                fp = open(path, 'rb')
            except OSError:
                return False, {}

            from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
            from pdfminer.converter import PDFConverter
            from pdfminer.pdfpage import PDFPage
            from pdfminer.layout import LTImage, LTPage, LTFigure

            run_ocr = self.ocr.run
            ocr_result = [False]

            class MyImageReader(PDFConverter):
                def __init__(self, output, rdmngr, **kwargs):
                    super().__init__(rdmngr, io.BytesIO(), **kwargs)
                    self.output = output
                    self.output.append(False)
                    self.output.append("")
                    self.output.append("")

                def receive_layout(self, ltpage):
                    if isinstance(ltpage, (LTFigure, LTPage)):
                        for obj in ltpage:
                            self.receive_layout(obj)
                    if isinstance(ltpage, LTImage):
                        try:
                            self.ocr_pdf_page(ltpage.stream)
                        except Exception as exc:
                            logging.error(f"[pdf] Could not run OCR on image from PDF: {exc}")

                def ocr_pdf_page(self, stream):
                    img_ = PIL.Image.open(io.BytesIO(stream.get_rawdata()))

                    result = run_ocr(img_)
                    if result.success:
                        self.output[0] = True
                        self.output[1] = result.language
                        self.output[2] += result.fulltext + "\n"

            # basic document info
            try:
                parser = PDFParser(fp)
                pdf = PDFDocument(parser)
            except:
                fp.close()
                return False, {}

            if self.should_ocr(path):
                # OCR every image
                rmngr = PDFResourceManager(caching=True)
                device = MyImageReader(ocr_result, rmngr)
                interpreter = PDFPageInterpreter(rmngr, device)
                try:
                    for page in PDFPage.get_pages(fp, set()):
                        interpreter.process_page(page)
                except Exception as exc:
                    logging.error(f"[pdf] OCR in PDF died: {exc}")

            result = multidict.MultiDict()

            if ocr_result[0]:
                result.add('ocr.language', ocr_result[1])
                if self.should_fulltext(path):
                    result.add('ocr.fulltext', ocr_result[2])

            if len(pdf.info) > 0:
                for field in pdf.info[0].keys():
                    if not isinstance(field, str):
                        logging.debug(f"PDF: Unexpected type for info field key: {type(field)}")
                        continue
                    if field.lower() not in self.PDF_METADATA:
                        logging.debug(f"PDF: Ignoring {field} key")
                        continue

                    raw = pdf.info[0][field]
                    value = None

                    if isinstance(raw, bytes):
                        value = to_utf8(raw)
                        if value is None:
                            continue
                    elif isinstance(raw, str) and len(raw.strip()) > 0:
                        value = raw.strip()
                    else:
                        continue

                    if field.endswith('Date') and value.startswith(':'):
                        try:
                            value = datetime.datetime.strptime(value[:15], ':%Y%m%d%H%M%S')
                        except ValueError:
                            continue

                    if value is not None:
                        result['pdf.' + field] = value

            fp.close()
            return True, result


@registered_indexer
class EpubIndexer(Indexer):
    NAME = 'epub'
    ACCEPT = ['application/epub+zip']

    def run(self, path, info):
        logging.debug(f"[epub] processing {path.name}")

        with zipfile.ZipFile(path) as fp:
            files = fp.namelist()
            if 'content.opf' in files:
                with fp.open('content.opf') as contentfp:
                    return True, parse_opf(contentfp.read(), '')
        return False, {}


@registered_indexer
class FileTagsIndexer(Indexer):
    NAME = 'filetags'
    ACCEPT = '*'

    TAG_MARKER = ' -- '
    DATE_PATTERN = (re.compile(r'^([0-9]+)-([01][0-9])-([0-3][0-9])'), 10)
    DATE2_PATTERN = (re.compile(r'^([0-9]+)([01][0-9])([0-3][0-9])'), 8)
    DATETIME_PATTERN = (re.compile(r'^([0-9]+)-([01][0-9])-([0-3][0-9])[T_]([0-2][0-9])[._-]([0-6][0-9])'), 16)
    DATETIME2_PATTERN = (re.compile(r'^([0-9]+)([01][0-9])([0-3][0-9])_([0-2][0-9])([0-6][0-9])'), 13)
    DATETIMESEC_PATTERN = (re.compile(r'^([0-9]+)-([01][0-9])-([0-3][0-9])[T_]([0-2][0-9])[._-]([0-6][0-9])[._-]([0-6][0-9])'), 19)
    DATETIMESEC2_PATTERN = (re.compile(r'^([0-9]+)([01][0-9])([0-3][0-9])_([0-2][0-9])([0-6][0-9])([0-6][0-9])'), 15)

    def run(self, path, info):
        logging.debug(f"[filetags] Running {path.stem}")
        result = set()
        counter = 0

        success, result = self.extract_metadata(path.stem)

        while path.parent != path:
            path = path.parent
            counter += 1

            if counter > 1:
                break

            success, tags = self.extract_metadata(path.stem)
            result |= {(tag, value) for tag, value in tags if tag in [self.NAME + '.date', self.NAME + '.tags']}

        return len(result) > 0, multidict.MultiDict(list(result))

    def extract_metadata(self, text):
        result = set()
        tags = None

        match, text = self.obtain_datetime(text)

        if match:
            result.add((self.NAME + '.date', match))

        if match and text.startswith('--'):
            _, text = self.obtain_datetime(text[2:])
            # TODO: do something useful with the end part of the range

        if self.TAG_MARKER in text:
            text, tags = text.split(self.TAG_MARKER, 1)
            result |= {(self.NAME + '.tags', tag) for tag in tags.split()}

        if tags is not None or match is not None:
            if text.startswith('-'):
                text = text[1:]
            result.add((self.NAME + '.title', text))

        return len(result) > 0, result

    def obtain_datetime(self, text):
        match = None
        patterns = [
            self.DATETIMESEC_PATTERN,
            self.DATETIMESEC2_PATTERN,
            self.DATETIME_PATTERN,
            self.DATE_PATTERN,
            self.DATE2_PATTERN,
        ]

        for pattern, length in patterns:
            if len(text) < length:
                continue

            match = pattern.match(text)
            if not match:
                continue

            try:
                match = datetime.datetime(*[int(value) for value in match.groups()])
                text = text[length:]
                break
            except ValueError:
                match = None
                continue

        return match, text




import pathlib
import unittest

from metaindex import configuration
from metaindex.main import collect_files_and_metadata

HERE = pathlib.Path(__file__).resolve().parent
BASEDIR = HERE.parent


class TestCollect(unittest.TestCase):
    def setUp(self):
        self.config = configuration.load(HERE / 'test.conf')

    def test_only_dir(self):
        paths, metadata = collect_files_and_metadata(self.config, [BASEDIR])
        self.assertEqual(len(paths), 1)
        self.assertIn(BASEDIR / ".metadata.json", metadata)

    def test_single_file(self):
        paths, metadata = collect_files_and_metadata(self.config, [BASEDIR / "README.md"])
        self.assertEqual(len(paths), 1)
        self.assertIn(BASEDIR / ".metadata.json", metadata)

    def test_expand_dir(self):
        paths, metadata = collect_files_and_metadata(self.config, [BASEDIR], expand=True)
        self.assertIn(BASEDIR / "README.md", paths)
        self.assertIn(BASEDIR / "setup.py", paths)
        self.assertNotIn(BASEDIR / "metaindex", paths)
        self.assertNotIn(BASEDIR / "metaindex" / "main.py", paths)
        self.assertIn(BASEDIR / ".metadata.json", metadata)

    def test_recursive(self):
        paths, metadata = collect_files_and_metadata(self.config, [BASEDIR], recursive=True)
        self.assertIn(BASEDIR / ".metadata.json", metadata)
        self.assertIn(BASEDIR / "metaindex" / ".metadata.json", metadata)
        self.assertIn(BASEDIR / "tests" / ".metadata.json", metadata)

    def test_expand_recursive(self):
        paths, metadata = collect_files_and_metadata(self.config, [BASEDIR], expand=True, recursive=True)
        self.assertIn(BASEDIR / "README.md", paths)
        self.assertIn(BASEDIR / "metaindex" / "main.py", paths)
        self.assertIn(BASEDIR / "tests" / "test_collect.py", paths)




import unittest

from metaindex.query import tokenize, Query


class TestTokenizer(unittest.TestCase):
    def test_simple_words(self):
        text = '\\(there\\) once "was a"  \\"lady (in riga )'
        expected = ['(there)', 'once', 'was a', '\"lady', '(in', 'riga', ')']
        self.assertEqual(expected, tokenize(text))


class TestQueryParser(unittest.TestCase):
    def test_simple(self):
        text = "look, albatross"
        query = Query.parse(text)

        self.assertEqual(2, len(query.root.terms))

    def test_keyvalue(self):
        text = "title:albatross"
        query = Query.parse(text)

        self.assertEqual(query.root.terms[0].key, "title")
        self.assertEqual(query.root.terms[0].regex, "albatross")




import collections
import datetime
import pathlib
import sqlite3
import os
import logging
import re
import sys
import fnmatch

import multidict

from metaindex import indexer
from metaindex import configuration
from metaindex import sql
from metaindex import stores
from metaindex import shared
from metaindex import ocr
from metaindex.query import Query


class Cache:
    CACHE_FILENAME = "index" + os.extsep + "db"

    TYPE_DATETIME = 'd'
    TYPE_INT = 'i'
    TYPE_FLOAT = 'f'
    TYPE_STR = None

    Entry = collections.namedtuple('Entry', ['path', 'mimetype', 'last_accessed'])

    def __init__(self, config=None):
        self.config = config or configuration.load()
        self.db = None

        self.collection_metadata = config.list('General', 'collection-metadata', '')
        self.collection_metadata = set(sum([[fn + store.SUFFIX for store in stores.STORES] for fn in self.collection_metadata], start=[]))
        self.recursive_extra_metadata = config.bool("General", "recursive-extra-metadata", "y")
        self.ignore_dirs = config.list('General', 'ignore-directories', "", separator="\n")
        self.ignore_tags = config.list('General', 'ignore-tags', "")

        ocr_opts = config.list("General", "ocr", "no")
        if len(ocr_opts) == 0 or ocr_opts[0].lower() in config.FALSE:
            ocr_opts = False
        elif ocr_opts[0].lower() in config.TRUE:
            ocr_opts = True

        if ocr_opts:
            self.ocr = ocr.TesseractOCR(ocr_opts)
        else:
            self.ocr = ocr.Dummy()

        fulltext_opts = config.list("General", "fulltext", "no")
        if len(fulltext_opts) == 0 or fulltext_opts[0].lower() in config.FALSE:
            fulltext_opts = False
        elif fulltext_opts[0].lower() in config.TRUE:
            fulltext_opts = True

        self.extract_fulltext = fulltext_opts

        self.ignore_file_patterns = None
        self.accept_file_patterns = None

        accept = config.list('General', 'accept-files', '', separator="\n")
        ignore = config.list('General', 'ignore-files', '', separator="\n")

        if len(accept) > 0:
            self.accept_file_patterns = [re.compile(fnmatch.translate(pattern.strip()), re.I) for pattern in accept]
        elif len(ignore) > 0:
            self.ignore_file_patterns = [re.compile(fnmatch.translate(pattern.strip()), re.I) for pattern in ignore]

        self._open_db()

    def refresh(self, paths=None, recursive=True, processes=None):
        """(Re-)Index all items found in the given paths or all cached items if paths is None.

        If any item in the list is a directory and recursive is True, all
        items inside that directory (including any all subdirectories) will
        be indexed, too.

        processes may be set to a number of processes that will be used in
        parallel to index the files. If left at None, there will be as many
        processes launched as CPUs are available.
        """
        if paths is None:
            paths = [row['path'] for row in self.db.execute("select `path` from `files`")]

        elif isinstance(paths, (set, tuple)):
            paths = list(paths)

        elif not isinstance(paths, list):
            paths = [paths]

        if len(paths) == 0:
            return multidict.MultiDict()

        paths = [pathlib.Path(path).expanduser().resolve() for path in paths]

        # filter out ignored directories
        paths = [path for path in paths if not any([ignoredir in path.parts for ignoredir in self.ignore_dirs])]

        dirs = [path for path in paths if path.is_dir()]
        files = {path for path in paths
                      if path.is_file() and not self.is_sidecar_file(path) and self._accept_file(str(path))}
        files |= {fn for fn in indexer.find_files(dirs, recursive)
                     if not self.is_sidecar_file(fn) and self._accept_file(str(fn))}

        # TODO: allow disabling of sidecar storage(s)
        # existing sidecar files per file to be indexed
        sidecar_files = self._find_sidecar_files(files)
        
        # existing collection sidecar files per folder
        collection_sidecar_files = self._find_collection_sidecar_files(files)
        # keep track of what collection sidecar files have been parsed already, so we can be lazy
        parsed_collection_sidecar_files = set()

        # files to process through indexing
        to_index = []
        # cache of last_modified dates for files
        last_modified = dict([(fn, shared.get_last_modified(fn)) for fn in sum(sidecar_files.values(), start=[])]
                           + [(fn, shared.get_last_modified(fn)) for fn in sum(collection_sidecar_files.values(), start=[])]
                           + [(fn, shared.get_last_modified(fn)) for fn in files])
        # cache of extra metadata values per file (absolute path)
        extra_metadata_cache = {}

        def match_collection_sidecar_files(fn):
            queue = [fn.parent]
            if self.recursive_extra_metadata:
                parent = fn.parent.parent
                # walk up the tree and add all directories into the queue
                while True:
                    queue.append(parent)
                    if parent == parent.parent:
                        break
                    parent = parent.parent
            return sum([collection_sidecar_files.get(path, []) for path in queue], start=[])

        to_ignore = set()

        # check what files have changed since the cache was created
        for file_ in files:
            if file_ in to_ignore:
                continue

            coll_sidecars = match_collection_sidecar_files(file_)
            relevant_files = [file_] + sidecar_files.get(file_, []) + coll_sidecars
            current_lm = max([last_modified[fn] for fn in relevant_files])

            query = "select `last_modified` from `files` where `path` = ?"
            row = self.db.execute(query, (str(file_), )).fetchone()

            if row is None:
                to_index.append(file_)
            else:
                cached_lm = sql.str_to_dt(row['last_modified'])

                if current_lm > cached_lm:
                    logging.debug(f"Cached version of {file_.name} from {cached_lm} is out of date ({current_lm})")
                    to_index.append(file_)
                    # ensure we remember the last_modified
                    last_modified[file_] = current_lm

        logging.info(f"Refreshing {len(to_index)} files")
        
        indexer_result = indexer.index_files(to_index, processes, self.ocr, self.extract_fulltext, self.config)

        # cache metadata from collection sidecar files
        for file_ in sum(collection_sidecar_files.values(), start=[]):
            extra_metadata_cache.update(self.parse_extra_metadata(file_))
            parsed_collection_sidecar_files.add(file_)

        # obtain all collection metadata files
        for file_, _, _ in indexer_result:
            # obtain collection metadata for this file
            for fn in match_collection_sidecar_files(file_):
                if fn in parsed_collection_sidecar_files:
                    continue
                extra_metadata_cache.update(self.parse_extra_metadata(fn))
                parsed_collection_sidecar_files.add(fn)

        metadata = multidict.MultiDict()

        # add metadata to cache
        for file_, success, info in indexer_result:
            # start with collection metadata ('*' and '**' entries)
            base_metadata = extra_metadata_cache.get(file_.parent, multidict.MultiDict()).copy()
            success = success or len(base_metadata) > 1

            # collection metadata specific to this file
            sidecar_info = extra_metadata_cache.get(file_, multidict.MultiDict()).copy()
            if len(sidecar_info) > 1:
                success = True
                base_metadata.extend(sidecar_info)

            # apply sidecar file contents
            for sidecar in sidecar_files.get(file_, []):
                sidecar_info = stores.get(sidecar)
                if len(sidecar_info) > 1:
                    success = True
                    base_metadata.extend(sidecar_info)

            # collect deletions
            delete_keys = {key_.split('.', 1)[1]
                           for key_, value in base_metadata.items()
                           if value is None and key_.startswith('extra.')}
            success = success or len(delete_keys) > 0

            # apply metadata from the file itself
            base_metadata.extend(info)

            for key_ in delete_keys | {shared.IS_RECURSIVE}:
                base_metadata.popall(key_, True)

            if success:
                self.insert(file_, base_metadata, last_modified[file_])
                metadata[str(file_)] = base_metadata

        return metadata

    def cleanup(self):
        """Find and remove all entries in the cache that refer to no longer existing files"""
        to_delete_ids = []

        for row in self.db.execute("select `id`, `path` from files"):
            path = pathlib.Path(row['path'])

            if not path.exists():
                to_delete_ids.append(row['id'])

        self.forget_ids(to_delete_ids)

    def _accept_file(self, path):
        if self.accept_file_patterns is not None:
            return any([pattern.match(path) for pattern in self.accept_file_patterns])
        return not any([pattern.match(path) for pattern in self.ignore_file_patterns])

    def clear(self):
        """Remove everything from the cache"""
        with self.db:
            self.db.execute("delete from metadata")
            self.db.execute("delete from files")

    def forget_ids(self, ids):
        """Delete these IDs from the cache"""
        with self.db:
            id_set = ", ".join(["?"]*len(ids))
            self.db.execute(f"delete from metadata where id in ({id_set})", ids)
            self.db.execute(f"delete from files where id in ({id_set})", ids)

    def expire_metadata(self, paths, recursive=True):
        """Remove all metadata associated to these paths

        But keep the paths in the database.
        """
        if paths is None:
            self.db.execute("update `files` set `last_modified` = ?", (sql.MIN_DATE,))
            return

        elif not isinstance(paths, list):
            paths = [paths]

        paths = [pathlib.Path(path).expanduser().resolve() for path in paths]

        query = 'update `files` set `last_modified` = ? where `path` in (' + \
                ", ".join(["?"]*len(paths)) + ")"
        self.db.execute(query, [sql.MIN_DATE] + [str(path) for path in paths])

    def forget(self, paths, recursive=True):
        """Remove all paths from the cache

        If recursive is True, all items in subdirectories are removed from the
        cache, too."""
        raise NotImplementedError()

    def find(self, query):
        """Find all items matching this query.

        query may either be a human-written search term or a Query instance.

        Returns a list to (path to file, metadata) tuples.
        """
        if query is None or len(query) == 0:
            query = Query()

        elif isinstance(query, str):
            query = Query.parse(query, self.config.synonyms)

        elif not isinstance(query, Query):
            raise TypeError()

        qry, args = query.as_sql()
        ids = [row['id'] for row in self.db.execute(qry, args)]
        
        return self._resolve_ids(ids)

    def get(self, paths, recursive=True):
        """Get metadata for all items of paths

        paths may also be a single path instead of a list of paths.

        If any element of paths is pointing to a directory and recursive is
        set to True, get will return the metadata for all elements inside
        that directory (and their subdirectories'), too.
        """
        if not isinstance(paths, list):
            paths = [paths]

        paths = [pathlib.Path(path).expanduser().resolve() for path in paths]
        
        args = [str(path) for path in paths if not path.is_dir()]
        if len(args) == 0:
            return []

        query = "select distinct `files`.`id` from `files` where `files`.`path` in (" \
                + ", ".join(["?"]*len(args)) + ")"

        ids = [row['id'] for row in self.db.execute(query, args)]

        return self._resolve_ids(ids)

    def keys(self):
        """Returns a set of all known metadata keys."""
        return {row['key'] for row in self.db.execute("select distinct `key` from `metadata`")}

    def insert(self, path, metadata, last_modified=None):
        """Insert the metadata for item at path into the cache.

        This operation will not modify the item in the filesystem nor update
        any other form of metadata persistency for the item.
        This function really only affects the cache.
        """
        # determine rowid, create new entry if there is none yet
        rowids = [row['id']
                 for row in self.db.execute("select `id` from `files` where `path` = ?",
                                            (str(path), ))]
        if last_modified is None:
            last_modified = shared.get_last_modified(path)

        if len(rowids) == 0:
            with self.db:
                result = self.db.execute("insert into `files`(`path`, `last_modified`) values(?, ?)",
                                         (str(path), sql.dt_to_str(last_modified)))
                rowid = result.lastrowid
            logging.debug(f"{path.name} not in cache yet. New ID is {rowid}")
        else:
            rowid = rowids[0]
            self.db.execute("update `files` set `last_modified` = ? where `id` = ?",
                            (sql.dt_to_str(last_modified), rowid))

        # clear existing metadata and write the current
        with self.db:
            self.db.execute("delete from `metadata` where `id` = ?", (rowid, ))

            for key in metadata.keys():
                if key in self.ignore_tags:
                    continue

                values = sum([value if isinstance(value, list) else [value]
                              for value in metadata.getall(key)
                              if value is not None], start=[])

                for value in values:
                    type_ = None
                    if isinstance(value, datetime.datetime):
                        type_ = Cache.TYPE_DATETIME
                        value = sql.dt_to_str(value)
                    elif isinstance(value, int):
                        type_ = Cache.TYPE_INT
                        value = str(value)
                    elif isinstance(value, float):
                        type_ = Cache.TYPE_FLOAT
                        value = str(value)
                    elif not isinstance(value, str):
                        logging.error(f"Unexpected type {type(value)} for key {key}, skipping")
                        logging.debug(f"Unexpected type {type(value)} for key {key} with value {value}")
                        continue
                    self.db.execute("insert into `metadata` values(?, ?, ?, ?)",
                                    (rowid, key, type_, value))

    def parse_extra_metadata(self, metafile):
        """Extract extra metadata from this file"""
        data = stores.get_for_collection(metafile)

        for filename in data.keys():
            data[filename][shared.IS_RECURSIVE] = data[filename][shared.IS_RECURSIVE] and \
                                                  self.recursive_extra_metadata
        
        return data

    def is_sidecar_file(self, path):
        """Determine whether the file at path is a sidecar file"""
        if not path.is_file:
            return False
        if path.name in self.collection_metadata:
            return True
        for store in stores.STORES:
            if path.suffix == store.SUFFIX and len([fn for fn in path.parent.iterdir() if fn.is_file() and fn != path and fn.stem == path.stem]) > 0:
                return True
        return False

    def _resolve_ids(self, ids):
        result = {}
        query = "select files.path, metadata.* from `files` " \
                " left join `metadata` on `files`.`id` = `metadata`.`id` " \
                " where `files`.`id` in (" + ", ".join(["?"]*len(ids)) + ")"
        for row in self.db.execute(query, ids):
            id_ = row['id']
            if id_ not in result:
                result[id_] = (row['path'], multidict.MultiDict())

            key, value = self._translate_row(row)
            if not isinstance(key, str):
                logging.error(f"Unexpected key type {type(key)}")
                key = str(key)
            result[id_][1].add(key, value)

        return [v for v in result.values()]

    def _translate_row(self, row):
        key, type_, value = row['key'], row['type'], row['value']
        
        if type_ == Cache.TYPE_DATETIME:
            value = sql.str_to_dt(value)
        elif type == Cache.TYPE_INT:
            value = int(value)
        elif type == Cache.TYPE_FLOAT:
            value = float(value)

        return key, value

    def _open_db(self):
        if self.db is not None:
            self.db.close()

        location = self.config.path('General', 'cache', configuration.CACHEPATH) / Cache.CACHE_FILENAME
        if not location.parent.exists():
            location.parent.mkdir(parents=True, exist_ok=True)

        self.db = sqlite3.connect(location)
        self.db.row_factory = sqlite3.Row
        self.db.create_function('REGEXP', 2, sql.regexp)

        with self.db:
            self.db.execute(sql.CREATE_FILE_TABLE)
            self.db.execute(sql.CREATE_META_TABLE)

    def _find_sidecar_files(self, files):
        """Find all sidecar files for these files

        Returns a dict with the mapping of path -> [sidecar files] containing
        all existing sidecar files"""
        sfiles = dict([(file_, [fn for fn in stores.sidecars_for(file_)
                                   if fn.exists()]) for file_ in files])
        return dict([pair for pair in sfiles.items() if len(pair[1]) > 0])

    def _find_collection_sidecar_files(self, files):
        """Find all collection sidecar files for these files

        Returns a dict with the mapping of path -> [sidecar files] containing
        all existing sidecar files"""
        queue = {f.parent for f in files}
        if self.recursive_extra_metadata:
            to_visit = set()
            # find all parent directories
            while len(queue) > 0:
                path = queue.pop()
                to_visit.add(path)
                if path.parent != path and path.parent not in to_visit:
                    queue.add(path.parent)
        sfiles = dict([(base, [base / fn for fn in self.collection_metadata
                                         if (base / fn).exists()])
                       for base in to_visit])
        return dict([pair for pair in sfiles.items() if len(pair[1]) > 0])


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    cache = Cache()
    breakpoint()




import codecs
import datetime
import pathlib
import mimetypes
import multiprocessing
import logging
import zipfile

import multidict

from metaindex import ocr
from metaindex import configuration


_registered_indexers = {}
_indexer_by_suffix = {}
_indexer_by_mimetype = {}
_generic_indexers = []

ocr_facility = ocr.Dummy()
app_config = configuration.Configuration()


class Order:
    FIRST = 100
    EARLY = 250
    AVERAGE = 500
    LATE = 750
    LAST = 1000


class Indexer:
    # specify what suffices or mimetypes are handled by this indexer, e.g.
    # ACCEPT = ['.rst', '.md', 'text/html', 'image/']
    # anything starting with a . is expected to be a suffix,
    # everything else is expected to be a mimetype.
    # If the mimetype ends with /, it is matched against the first part
    # of the file's mimetype
    # If your indexer should run for all files, use ACCEPT = '*'
    ACCEPT = []
    # When to execute this indexer in the order of indexers
    ORDER = Order.AVERAGE

    def __init__(self, ocr, fulltext, config):
        """Create a new indexer with this configuration

        ocr: the OCR facility to use (e.g. metaindex.ocr.Dummy)
        fulltext: whether or not to extract the fulltext
        """
        self.ocr = ocr
        self.fulltext = fulltext
        self.config = config

    def should_ocr(self, path):
        return self._check_match(self.ocr.accept_list, path)

    def should_fulltext(self, path):
        return self._check_match(self.fulltext, path)

    def _check_match(self, acceptor, path):
        if not isinstance(acceptor, list):
            return bool(acceptor)

        if self.NAME in acceptor:
            return True

        for item in acceptor:
            if item.startswith('.') and path.suffix.lower() == item.lower():
                return True
            if '/' in item:
                mimetype, _ = mimetypes.guess_type(path.name, strict=False)
                if mimetype.lower().startswith(item.lower()):
                    return True

        return False

    def __lt__(self, other):
        return self.ORDER < other.ORDER

    def run(self, path, info):
        """Execute this Indexer to run on the file at path.

        ``info`` is the previously collected metadata as a multidict this is
        purely informative and should not be returned.

        Return the tuple (success, metadata). success being True or False,
        metadata being a multidict.Multidict or a dict.

        path will be of type pathlib.Path.

        Be aware that each subprocess will create their own instance of the
        Indexer.
        """
        raise NotImplementedError()


def registered_indexer(cls):
    assert issubclass(cls, Indexer)
    
    global _registered_indexers
    global _indexer_by_suffix
    global _indexer_by_mimetype
    global _generic_indexers

    assert cls.NAME not in _registered_indexers
    _registered_indexers[cls.NAME] = cls

    if isinstance(cls.ACCEPT, str) and cls.ACCEPT == '*':
        _generic_indexers.append(cls.NAME)
    else:
        for pattern in cls.ACCEPT:
            pattern = pattern.lower()

            if pattern.startswith('.') and len(pattern) > 1:
                if pattern not in _indexer_by_suffix:
                    _indexer_by_suffix[pattern] = []
                _indexer_by_suffix[pattern].append(cls.NAME)

            elif len(pattern) > 0:
                if pattern.endswith('/') and len(pattern) > 1:
                    pattern = pattern[:-1]
                if pattern not in _indexer_by_mimetype:
                    _indexer_by_mimetype[pattern] = []
                _indexer_by_mimetype[pattern].append(cls.NAME)

    return cls


def remove_indexers(names):
    global _registered_indexers
    global _indexer_by_suffix
    global _indexer_by_mimetype
    global _generic_indexers

    if len(names) == 0:
        return

    for name in names:
        if name in _registered_indexers:
            del _registered_indexers[name]

    _indexer_by_suffix = dict([(key, [value for value in values if value not in names])
                               for key, values in _indexer_by_suffix.items()])
    _indexer_by_mimetype = dict([(key, [value for value in values if value not in names])
                                 for key, values in _indexer_by_mimetype.items()])
    _generic_indexers = [value for value in _generic_indexers if value not in names]


def to_utf8(raw):
    if isinstance(raw, str):
        return raw
    encoding = None
    skip = 1

    if raw.startswith(codecs.BOM_UTF8):
        encoding = 'utf-8'
    elif raw.startswith(codecs.BOM_UTF16_BE):
        encoding = 'utf-16-be'
    elif raw.startswith(codecs.BOM_UTF16_LE):
        encoding = 'utf-16-le'
    elif raw.startswith(codecs.BOM_UTF32_BE):
        encoding = 'utf-32-be'
    elif raw.startswith(codecs.BOM_UTF32_LE):
        encoding = 'utf-32-le'
    else:
        # just best efford
        encoding = 'utf-8'
        skip = 0

    try:
        text = str(raw, encoding=encoding).strip()
        return text[skip:]  # drop the BOM, if applicable
    except UnicodeError:
        pass
    return None


class IndexerCache:
    def __init__(self, ocr, fulltext, config):
        self.cached = {}
        self.ocr = ocr
        self.fulltext = fulltext
        self.config = config

    def get(self, name):
        if name not in self.cached:
            self.cached[name] = _registered_indexers[name](self.ocr, self.fulltext, self.config)
        return self.cached[name]


def get_metadata(filename, indexer_cache):
    """Extract metadata from the file at `filename`"""
    assert isinstance(filename, pathlib.Path)

    global ocr_facility
    global obtain_fulltext
    global app_config

    logging.error(f"Going for {filename}")

    if indexer_cache is None:
        indexer_cache = IndexerCache(ocr=ocr_facility,
                                     fulltext=obtain_fulltext,
                                     config=app_config)

    stat = filename.stat()
    suffix = filename.suffix[1:]
    mimetype, _ = mimetypes.guess_type(filename, strict=False)
    info = multidict.MultiDict({'size': stat.st_size,
                                'filename': filename.name,
                                'last_accessed': datetime.datetime.fromtimestamp(stat.st_atime),
                                'last_modified': datetime.datetime.fromtimestamp(stat.st_mtime)})

    if mimetype is None:
        logging.debug(f"Unknown mimetype for {suffix}")
        return False, info

    info['mimetype'] = mimetype
    applied_indexers = 0
    indexers = _generic_indexers[:] \
             + _indexer_by_suffix.get(filename.suffix.lower(), [])
    for mtype in [mimetype, mimetype.split('/', 1)[0]]:
        indexers += _indexer_by_mimetype.get(mtype, [])

    for handler in sorted([indexer_cache.get(indexer) for indexer in indexers]):
        logging.error(f"... running {handler}")
        success, extra = handler.run(filename, info.copy())
        if success:
            applied_indexers += 1
            info.extend(extra)

    success = applied_indexers > 0
    return success, info


def indexer(filenames):
    """Takes a list of filenames and tries to extract the metadata for all
    
    Returns a dictionary mapping filename to a dictionary with the metadata.
    """
    global ocr_facility
    global obtain_fulltext
    global app_config

    indexer_cache = IndexerCache(ocr_facility, obtain_fulltext, app_config)
    for name in _registered_indexers.keys():
        # pre-load all
        indexer_cache.get(name)

    result = []

    for filename in filenames:
        if not isinstance(filename, pathlib.Path):
            filename = pathlib.Path(filename)

        if not filename.is_file():
            continue

        success, info = get_metadata(filename, indexer_cache)

        result.append((filename, success, info))

    return result


def find_files(paths, recursive=True):
    """Find all files in these paths"""
    if not isinstance(paths, list):
        paths = [paths]

    pathqueue = list(paths)
    filenames = []

    while len(pathqueue) > 0:
        path = pathqueue.pop(0)

        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        if not path.exists():
            continue

        for item in path.iterdir():
            if item.is_dir() and recursive:
                pathqueue.append(item)
                continue

            if item.is_file():
                filenames.append(item)

    return filenames


def tuplice(item):
    success, info = get_metadata(item, None)
    return item, success, info


def index_files(files, processes=None, ocr=None, fulltext=False, config=None):
    """Run indexer on all files"""
    global ocr_facility
    global obtain_fulltext
    global app_config

    if ocr is not None:
        ocr_facility = ocr
    app_config = config or configuration.Configuration()
    obtain_fulltext = fulltext

    metadata = {}
    then = datetime.datetime.now()
    if processes == 1:
        index_result = indexer(files)
    else:
        with multiprocessing.Pool(processes) as pool:
            index_result = pool.map(tuplice, files)

    logging.info(f"Processed {len(metadata)} files in {datetime.datetime.now() - then}")

    return index_result


if __name__ == '__main__':
    import sys

    logging.basicConfig(level=logging.DEBUG)

    index = index_files([pathlib.Path(i).expanduser() for i in sys.argv[1:]])

